//******************************************************************************
//Name          : SystemIdentification.cs
//Description   : Implementation to experience the beauty and the pitfalls of System
//                Identification
//Author        : Dr. Oliver Kluge, 90411 Nuernberg, Germany
//Copyright(C)  : Oliver Kluge 2014 
//------------------------------------------------------------------------------
//******************************************************************************

using System;
using Display;


class Identifikation
{
    const int samplerate = 1024;                    // Abtastwerte im Abtastintervall
    const double PI = 3.141592653589793;

    public double[] x  = new double[samplerate];    // Eingangssignal im Zeitbereich
    public double[] y  = new double[samplerate];    // Ausgangssignal im Zeitbereich
    public double[] arg = new double[samplerate];
    public double[] XA = new double[samplerate];    // Eingangssignal im Frequenzbereich
    public double[] XP = new double[samplerate];
    public double[] YA = new double[samplerate];    // Ausgangssignal im Frequenzbereich
    public double[] YP = new double[samplerate];
    public double[] HA = new double[samplerate];    // Amplitudengang des bertragungssystems
    public double[] HP = new double[samplerate];

    public int Samplerate
    {
        get { return samplerate; }
    }

    public void Filterung(int select)
    {
        int M = 10;
        double[] a = new double[M];

        // Mittelwertfilter als "unbekanntes" bertragungssystem
        if (select == 1)
        {
            for (int k=0; k<M; k++)
            {
                a[k] = 1.0/(double)M;
            }

            for (int k=M-1; k<samplerate; k++)
            {
                for (int j=0; j<M; j++)
                {
                    y[k] += x[k-j] * a[j];
                }
            }
        }

        // IIR-Tiefpassfilter 1. Ordnung als "unbekanntes" bertragungssystem
        if (select == 2)
        {
            for (int k=1; k<samplerate; k++)
            {
                y[k] = x[k] + 0.5*y[k-1];
            }
        }

        // IIR-Filter 2. Ordnung als "unbekanntes" bertragungssystem
        if (select == 3)
        {
            for (int k=2; k<samplerate; k++)
            {
                y[k] = x[k] + y[k-1] - 0.5*y[k-2];
            }
        }

        // IIR-Filter 4. Ordnung (2*2 kaskadiert) als "unbekanntes" bertragungssystem
        // y[k] = x[k] + y[k-1] - 0.5*y[k-2] kaskadiert
        if (select == 4)
        {
            for (int k=4; k<samplerate; k++)
            {
                y[k] = x[k] + 2*y[k-1] - 2*y[k-2] + y[k-3] - 0.25*y[k-4];
            }
        }
    }

    public void CreateTestsignal(int select)
    {
        // Eingangssignal = Einheitsimpuls
        if (select == 1)
        {
            for (int k=0; k<samplerate; k++)
            {
                if (k == samplerate/2)
                {
                    x[k] = 1;
                }
                else
                {
                    x[k] = 0;
                }
            }
        }

        // Eingangssignal = Sprungfunktion
        if (select == 2)
        {
            for (int k=0; k<samplerate; k++)
            {
                if (k < samplerate/2)
                {
                    x[k] = 0;
                }
                else
                {
                    x[k] = 1;
                }
                // zustzliche Hanning-Fensterung bei Sprungfkt.
                x[k] *= (0.5 - 0.5*Math.Cos(2*PI*k/samplerate));
            }
        }

        // Eingangssignal = weisses Rauschen
        if (select == 3)
        {
            MyRandom.Random noise = new MyRandom.Random(0.71036, 0.23906);

            for (int k=0; k<samplerate; k++)
            {
                x[k] = noise.Normal(0, 1);
            }
        }

        // Eingangssignal = Gleitsinus
        if (select == 4)
        {
            double f = 0;
            for (int k=0; k<samplerate; k++)
            {
                x[k] = Math.Sin(2*PI*f*k/samplerate);
                f += 0.25;
            }
        }

        // store argument data
        for (int k=0; k<samplerate; k++)
        {
            arg[k] = k;// 2 * PI * k / samplerate;
        }
    }

    public void Frequenzgang(double StartFreq, double StopFreq)
    {
        double deltaf = (StopFreq-StartFreq)/(samplerate-1);
        double freq   = StartFreq;
        double real   = 0;
        double imag   = 0;

        // ber alle Frequenzen
        for (int i=0; i<samplerate; i++)
        {
            real = 0;
            imag = 0;

            // ber alle Abtastwerte
            for (int k=0; k<samplerate; k++)
            {
                real += x[k]*Math.Cos(2*PI*freq*k/samplerate); // an
                imag += x[k]*Math.Sin(2*PI*freq*k/samplerate); // bn
            }
            // DFT des Eingangssignals
            XA[i] = Math.Sqrt(real*real + imag*imag);
            XP[i] = Math.Atan2(imag, real + 1.0E-10);

            real = 0;
            imag = 0;

            // ber alle Abtastwerte
            for (int k=0; k<samplerate; k++)
            {
                real += y[k]*Math.Cos(2*PI*freq*k/samplerate);
                imag += y[k]*Math.Sin(2*PI*freq*k/samplerate);
            }
            // DFT des Ausgangssignals
            YA[i] = Math.Sqrt(real*real + imag*imag);
            YP[i] = Math.Atan2(imag, real + 1.0E-10);

            // Amplitudengang des bertragungssystems H(f)=Y(f)/X(f)
            HA[i] = YA[i]/(XA[i] + 1.0E-10);	// Division durch Null vermeiden!!!
            HP[i] = YP[i] - XP[i];

            freq += deltaf;
        }
    }
}


class MainApp
{
    [STAThread]
    static void Main()
    {
        string userinput = "";
        int select1 = -1;
        int select2 = -1;
        Identifikation mySystem = new Identifikation();

        Console.Write("Bestimmung des Frequenzgangs eines unbekannten bertragungssystems\n");
        Console.Write("mit Hilfe verschiedener Testfunktionen:\n\n");

        Console.Write("zu testendes bertragungssystem:\n");
        Console.Write("Mittelwertfilter mit M=10                         ->  1\n");
        Console.Write("IIR-Tiefpass 1. Ordnung                           ->  2\n");
        Console.Write("IIR-Tiefpass 2. Ordnung mit konj.kompl. Polen     ->  3\n");
        Console.Write("IIR-Filter 4. Ord. 2*2 kask. mit kon.kompl. Polen ->  4\n");
        userinput = Console.ReadLine();
        select1 = Convert.ToInt32(userinput);

        Console.WriteLine("\n");
        Console.Write("verwendete Testfunktion:\n");
        Console.Write("Einheitsimpuls     ->  1\n");
        Console.Write("Einheitssprung     ->  2\n");
        Console.Write("weies Rauschen    ->  3\n");
        Console.Write("Gleitsinus (Chirp) ->  4\n");
        Console.WriteLine();
        userinput = Console.ReadLine();
        select2 = Convert.ToInt32(userinput);

        mySystem.CreateTestsignal(select2);
        mySystem.Filterung(select1);
        mySystem.Frequenzgang(0, mySystem.Samplerate/2);

        // create a chart for displaying that data...
        DisplayChart mydisplay1 = new DisplayChart(mySystem.arg, mySystem.x, "Time domain", "k", "x(k)");
        mydisplay1.Plot();

        DisplayChart mydisplay2 = new DisplayChart(mySystem.arg, mySystem.XA, "Frequency domain", "f", "|X(f)|");
        mydisplay2.Plot();

        DisplayChart mydisplay3 = new DisplayChart(mySystem.arg, mySystem.y, "Time domain", "k", "y(k)");
        mydisplay3.Plot();

        DisplayChart mydisplay4 = new DisplayChart(mySystem.arg, mySystem.YA, "Frequency domain", "f", "|Y(f)|");
        mydisplay4.Plot();

        DisplayChart mydisplay5 = new DisplayChart(mySystem.arg, mySystem.HA, "Frequency domain", "f", "|H(f)|");
        mydisplay5.Plot();
    }
}
